import java.io.*;
import java.util.*;
import java.lang.*;
import java.rmi.*;
import java.rmi.server.*;
import java.util.concurrent.PriorityBlockingQueue;
import java.net.*;

/** This class represents a physical processor involved in the time warp 
 * simulation engine. A PE could have multiple LPs running on it.
  * @author Yin Xiong
  * @version 2.0 (10/2006)
  */
public class PE implements Runnable{

  /**an array of Peer objects serving as the handles of the remote methods*/  
  Peer[] peers;
 
  /**The id of this PE.*/
  int myID;

  /**The name of the machine where the PE is running*/
  String myMachineName;

  /**The ip of the machine where the PE is running*/
  String myIP;

  /**The LP configuration information*/
  Vector<LPConfigInfo> lpConfigInfo;

  /**The most recent GVT*/
  int GVT;

  /**Sleep time*/
  int sleepTime;

 /**The global name server*/
  GNSInterface gns;
  
  /**The LocalNameServer in charge of routing messages.*/
  LocalNameServer localNameServer;
  

  /**The queue that holds messages that have been cancelled. When a processor
    * wishes to cancel a message, it enqueues the message being cancelled into
    * this queue. Logically, each message enqueued in this queue can be viewed
    * as an anti-message, however, it is the message itself rather than an
    * explicit anti-message that is enqueued. Access this queue is synchronized
    * with locks. 
    */
  PriorityBlockingQueue<Message> cancelQ;

  
 /**The queue that holds the outgoing messages. Messages are placed into this
    *by the Scheduler and the sender(main thread) will come here to get the
    *outgoing message and send it out. A copy is kept in this queue, but its sign
    *is changed from positive to negative.
    */
  PriorityBlockingQueue<Message> sendQ;
    
/**The queue that holds all unprocessed events for all LPs mapped to this PE.  
 * A single queue is used here in order to eliminates the need for a
 * separate "scheduling queue" to enumerate the executable LPs.
  */
  //Queue calendarQ;
  

  /**The queue that holds the events whose memeory is free to reclaim.*/
  PriorityBlockingQueue<Message> freeQ;

  /**The scheduler that schedules events for this PE.*/
  SchedulingLoop schedulingLoop;
  
  /**Flag for end of sim*/
  Flag flag;

  /**the name of the GNS server*/
  String serverName;

public PE(String sv, int numPEs)
{
     GVT=0;
     flag=new Flag();
     peers=new Peer[numPEs];
     sleepTime=0;
     lpConfigInfo=new Vector<LPConfigInfo>();
     localNameServer=new LocalNameServer(myID);
     cancelQ=new PriorityBlockingQueue<Message>();
     sendQ=new PriorityBlockingQueue<Message>();
     schedulingLoop=new SchedulingLoop(flag, myID);

  try{
     InetAddress address = InetAddress.getLocalHost();
     myMachineName=address.getHostName();
     myIP=address.getHostAddress();
     serverName="rmi://" + sv+ "/GNS";
   }
  catch(Exception e){
    e.printStackTrace();
  }

}//end constructor

/**Sets the end of sim flag so the PE will stop.
  */
void setEOSflag()
{
   flag.set(-1);
}//end setEOSflag

/**Connects this PE to the GNS server, receives LP information and create LPs
  *accordingly.
  */
void connectToGNS()throws RemoteException
{
      
    try{
        gns =(GNSInterface)Naming.lookup(serverName);
        lpConfigInfo=gns.registerPE(myMachineName);
        System.out.println(myMachineName + " connected to " + serverName 
                           + " #lp: " + lpConfigInfo.size());        
        

        LPConfigInfo lpinfo;
        if(lpConfigInfo.size()>0){
          lpinfo=(LPConfigInfo)lpConfigInfo.get(0);
          myID=lpinfo.PEid;
          schedulingLoop=new SchedulingLoop(flag,myID);
       
          //create LPs
          for(int i=0;i<lpConfigInfo.size();i++){
             lpinfo=(LPConfigInfo)lpConfigInfo.get(i);
             createLP(lpinfo);
           }//end for
         }//end if        
       }//end try
      catch(Exception e){
        e.printStackTrace();
      }

}//end connectToGNS

 /**Runs the PE(processing events).
   *This is the communication thread that send and receive messages.
   */

public void run()
   {
    System.out.println("PE " + myID + " is running ...\n");
    initLPs();   
	
    Thread scheduler=new Thread(schedulingLoop);
    scheduler.start();

    boolean endOfSim=false;

  while(!endOfSim){
   try{
    //send messages stored in sednQ
    rmiSendMessages();

    //process the queues
    processQs();

    if(flag.get()<0){
     //System.out.println("PE " + myID + " gets flag -1");
       endOfSim=true;
      }

    Thread thread=Thread.currentThread();
    thread.sleep(sleepTime);
  } 
  catch(Exception e){
    e.printStackTrace();
  }

 }//end while

    System.out.println(myMachineName + " finished running ");   
    finalizeLPs();
    
    finalizePE();     

}//end run


/**Finalizes the PE by checking and setting the end-of-sim flag.
  */
void finalizePE()
{
  try{
   int f=gns.getEOSflag();
   //System.out.println(myMachineName + " get flag from gns: "+ f);
   if(f>=0)
     gns.setEOSflag(-1);
     //System.out.println(myMachineName + " set gns flag to -1 ");

   }
  catch(Exception e){
     e.printStackTrace();
  }
   
}//end finalizePE

/**Sends messages by invoking the remote method receiveMsg(Message msg) on
  *the receiver machine.
  */
void rmiSendMessages()
{
  //System.out.println(myMachineName + " in send msg,sendQ size: "+sendQ.size());
  String dest=null; 

  for(int i=0;i<sendQ.size();i++){
      Message msg=(Message)sendQ.poll();
      int pid=msg.getReceiverPE();
      while(peers[pid]==null){      
         try{
            dest=msg.getReceiverMN();
            String targetName="rmi://" + dest + "/Peer";
            peers[pid] =(Peer)Naming.lookup(targetName);
           }
         catch(Exception e){
            //e.printStackTrace();
            System.out.println(myMachineName+ " trying to connect to " + dest);
          }
         doYield();  
      }//end while

      //System.out.println("connected to " + dest);
 
      if(peers[pid]!=null){
        int ack=-1;
        try{ 
          while(ack<0){
            ack=peers[pid].receiveMsg(msg);
            Thread thread=Thread.currentThread();
            thread.sleep(300);
           }//end while
        }
        catch(Exception e){
          e.printStackTrace();
        }  
     }
   }//end for
}//end rmiSendMessages


/**Processes events. It takes out each message in SchedulingLoop's outgoing
  *queue, finds out the routing information by application-specified id of the
  *destination or the LP id of the destination, consults the local name server
  *or the global name server (in case local name server does not have the info)
  *for the routing information. It then puts the message into the sendQ for 
  *sending out.
  */
void processEvents()
{
  boolean endOfSim=false;
     
   for(int i=0;i<schedulingLoop.outgoing.size();i++){
      Message msg=(Message)schedulingLoop.outgoing.poll();
      if(msg!=null){
         String sname=msg.getSenderAPPid();
         String rname=msg.getReceiverAPPid();
         System.out.println("PE " + myID + " sender,receive: " + sname+ " " +rname);

        RoutingInfo rinfo=null; 
        boolean byName=false;
        int lid=-1; 
        
       if(rname!=null){ 
            //check lns for receiver lp
             rinfo=localNameServer.getRoutingInfoByAid(rname);
             byName=true;
           }
        else{
            lid=msg.getReceiverLP();
            rinfo=localNameServer.getRoutingInfoByLid(lid);
        }       
        
        //System.out.println("PE " + myID + " get routing info from LNS: " + rinfo);
     
        //if lns doesn't have the receiver info, check gns
        while(rinfo==null){
          //System.out.println("gns: " + gns);
         try{
            if(byName==true)
                rinfo=gns.getRoutingInfoByAid(rname);
            else
                rinfo=gns.getRoutingInfoByLid(lid); 
          }
          catch(Exception e){
             e.printStackTrace();
          } 
          doYield();
        }//end while

       //System.out.println("PE " + myID + " get routinfo: " + rinfo);
       if(rinfo!=null){
          //store rinfo in local name server for future use
          localNameServer.addRoutingInfo(rinfo);
          msg.setReceiverPE(rinfo.PEid);
          msg.setReceiverAPPid(rinfo.APPid);
          msg.setReceiverLP(rinfo.LPid);
          msg.setReceiverMN(rinfo.machineName);

         //if the receiver is not on this machine, put the msg in sendQ
         if(rinfo.PEid!=myID){
           sendQ.add(msg); 
           //System.out.println("PE " + myID + " in proc events, sendQ size: " + sendQ.size());
         }//end if receiver is not on this machine

       else
          schedulingLoop.incoming.offer(msg);
   
      }//end if rinfo is not null
    
     }//end if msg!=null
  }//end for     
}//end ProcessEvents

/**The main function for this PE*/  
public static void main (String[] args)throws IOException {
  
String sv=null;
int numPEs=0;

if(args.length<2){
  System.out.println("Usage: java PE [GNS server name][number of PEs]");
  System.exit(1);
}

else{
  sv=args[0];
  numPEs=Integer.parseInt(args[1]);
}

  PE pe= new PE(sv,numPEs);
  pe.connectToGNS();

  MsgReceiver msgReceiver=new MsgReceiver(pe.schedulingLoop.incoming, pe.flag);

  Thread receiverThread=new Thread(msgReceiver);
  receiverThread.start();      
  pe.run();

}//end main


 /**Sets the sleep time for this thread; sleep time is specified by the user at the
   *command line. If not specified, it is set to 0
   */
public void setSleepTime(int st)
{
   sleepTime=st;
}


/**Yields the current thread.
  */
public void doYield()
{   
  try{
    Thread thread=Thread.currentThread();
    thread.yield();
  }
  catch(Exception e){
    e.printStackTrace();
  }
}//end doYield
     


/**Puts the current thread to sleep for sleeptime miliseconds.
  *@param sleeptime int the length for the sleep in miliseconds.
  */
void doSleep(int sleeptime)
{
  try{
    Thread thread=Thread.currentThread();
    thread.sleep(sleeptime);
  }
  catch(Exception e){
    e.printStackTrace();
  }        
}//end doSleep

  

/**Computes the local minimum timestamp
  *@return the local minimum timestamp
  *<p>local minimum timestamp = the minumum timestamp among
  *<ul>
  *<li>all unprocessed events
  *<li>all unacknowledged messages and anti-messages it has sent
  *<li>all marked acknowledgement messages it has received since it last received
  *    a new GVT value
  *</ul>
  * the PE sets a flag indicating it's in find mode
  */

int computeLocalMin()
{

  int localMin=-1;
  int temp;

  //check unprocessed event queue
  if(schedulingLoop.simMessages.size()>0){
    Message msg=(Message)schedulingLoop.simMessages.peek();
    temp=msg.timeStamp;
    if(temp>GVT){
       if(localMin<0 || (localMin>0 && temp<localMin))
         localMin=temp;
    }//end if 
  }//end if

 System.out.println("PE " + myID + " simMSG min: " + localMin);

  //check all marked acknowleged messages it has received since it last
  //received a new GVT
  for(int i=0;i<schedulingLoop.antiMsgsOut.size();i++){
   Message  msg=(Message)schedulingLoop.antiMsgsOut.get(i);
   temp=msg.timeStamp;
   //System.out.println("anti msg tag: " + msg.tag);
   if(msg.tag==1 && temp>GVT){
       if(localMin<0 || (localMin>0 && temp<localMin))
          localMin=temp;
    }//end if 
 }//end for

 System.out.println("PE " + myID + " antiMsgsOut min: " + localMin);


  //check anti-messaes it has received but not matched with an unprocessed message
  for(int i=0;i<schedulingLoop.antiMsgsIn.size();i++){
   Message  msg=(Message)schedulingLoop.antiMsgsIn.get(i);
   temp=msg.timeStamp;
   if(temp>GVT){
       if(localMin<0 || (localMin>0 && temp<localMin))
         localMin=temp;
      }//end if
  }//end for

 System.out.println("PE " + myID + " antiMsgsIn min: " + localMin);


 //check LP's timestamp
  for(int i=0;i<schedulingLoop.LPs.size();i++){
     LogicalProcess lp=(LogicalProcess)schedulingLoop.LPs.get(i);
     if(localMin<0 || (localMin>0 && lp.currentTime<localMin))
        localMin=lp.currentTime;
  }//end for

 System.out.println("PE " + myID + " lp min: " + localMin);

  return localMin;
}//end computeLocalMin


/**Updates GVT and does fossil collection.
  *@param newGVT int the new GVT broadcast by the controller
  */
void updateGVT(int newGVT)
{
   
   System.out.println("\nPE " + myID + " update GVT: old GVT, new GVT: " + GVT + " " + newGVT);  
   GVT=newGVT;
   int temp;

   //fossil collect all the space before the new GVT

  //fossil collection for unprocessed message
  if(schedulingLoop.simMessages.size()>0){
   for(int i=0;i<schedulingLoop.simMessages.size();i++){
      Message msg=(Message)schedulingLoop.simMessages.peek();
      temp=msg.timeStamp;
      if(temp<=GVT)
          schedulingLoop.simMessages.poll();
     }//end for
  }//end if


  //fossil collect the space from anti msges it sent
  for(int i=0;i<schedulingLoop.antiMsgsOut.size();i++){
   Message  msg=(Message)schedulingLoop.antiMsgsOut.get(i);
   temp=msg.timeStamp;
   if(temp<=GVT){
       schedulingLoop.antiMsgsOut.removeElementAt(i);
       System.out.println("PE " + myID + " remove antimsg out " + temp);
       i=i-1;
     }
  }//end for


  //fossil collect the space from anti msges it received
  for(int i=0;i<schedulingLoop.antiMsgsIn.size();i++){
   Message  msg=(Message)schedulingLoop.antiMsgsIn.get(i);
   temp=msg.timeStamp;
   if(temp<=GVT){
       schedulingLoop.antiMsgsIn.removeElementAt(i);
       System.out.println("PE " + myID + " remove antiMsgsIn " + temp);
       i=i-1;
     }
  }//end for

  //fossil collection for the LP states

  System.out.println("before update, PE " + myID + " has " + schedulingLoop.lpStates.size() + " states");
  for(int i=0;i<schedulingLoop.lpStates.size();i++){
    LogicalProcess lp=(LogicalProcess)schedulingLoop.lpStates.get(i);
    System.out.print (lp.currentTime + ", " );
   }//end for
   System.out.println();


  for(int i=0;i<schedulingLoop.lpStates.size();i++){
    LogicalProcess lp=(LogicalProcess)schedulingLoop.lpStates.get(i);
    if(lp.currentTime<=GVT){
       System.out.println("PE " + myID + " remove lpstate " + lp.currentTime);
       schedulingLoop.lpStates.removeElementAt(i);
       i=i-1;
      }//end if 
  }//end for

  System.out.println("\nafter update, PE " + myID + " has " + schedulingLoop.lpStates.size() + " states");
  for(int i=0;i<schedulingLoop.lpStates.size();i++){
    LogicalProcess lp=(LogicalProcess)schedulingLoop.lpStates.get(i);
    System.out.print (lp.currentTime + ", " );
   }//end for
   System.out.println();

}//end updateGVT



/**Processes the queues.
  */

void processQs()
{
    int cqsize,eqsize;
    cqsize=eqsize=0;

    //process events
    eqsize=schedulingLoop.outgoing.size();
    if(eqsize>0)
       processEvents();
   
    cqsize=cancelQ.size();
    if(cqsize>0)
      processCanQ();
      
}//end processQs


/**Places AntiMessages in CanQ.  These are events that have been scheduled by other events.
  *They are removed and deleted.
  */
void processCanQ()
{
     //System.out.println("PE " + myID + " is processing can queue...");
      
}//end ProcessCanQ
        


/**Create LPs according to the LP configuration information obtained from the
  *GNS(Global Name Server).
  *@param lpinfo LPConfigInfo object containing the lp configuration information
  */

void createLP(LPConfigInfo lpinfo)
{
   int pid=lpinfo.PEid;
   int lid=lpinfo.LPid;
   String appid=lpinfo.APPid;
   String appName=lpinfo.APPname;
   String className=lpinfo.className;
   String dat=lpinfo.data;
   schedulingLoop.createLP(pid,lid,appid,appName,className,dat);
}//end createLPs


/**Gets the LP by its id.
  *@param lpid int the LP id.
  */
LogicalProcess getLPByID(int lpid)
{
//System.out.println("in getLPByID,size:" + LPs.size());

  for(int i=0;i<schedulingLoop.LPs.size();i++){
    LogicalProcess lp=(LogicalProcess)schedulingLoop.LPs.get(i);
    if(lp.LPid==lpid)
      return lp;
  }  
  return null;
}//end getLPByID  


/**Initialize LPs by invoking initializeLP method provided by subclasses 
  *of LogicalProcess.
  */
void initLPs()
{
   for(int i=0;i<schedulingLoop.LPs.size();i++){
     LogicalProcess sublp=(LogicalProcess)schedulingLoop.LPs.get(i);
     sublp.initializeLP();
  }//end for     
      
}//end initLPs

/**Finalizes LPs by invoking finalizeLP method provided by subclasses of
  *LogicalProcess.
  */
void finalizeLPs()
{

   for(int i=0;i<schedulingLoop.LPs.size();i++){
     LogicalProcess sublp=(LogicalProcess)schedulingLoop.LPs.get(i);
     sublp.finalizeLP();
  }//end for     
  
}//end finalizeLPs


  /**Displays LP's publishing and subscribing attributes.*/
  /*
  public void displayInterests()
  {
   System.out.println("\nPE: " + myID);
    for(int i=0;i<LPs.size();i++){
	   LogicalProcess lp=(LogicalProcess)LPs.get(i);
	   lp.displayInterests();
	}//end for
  }//end displayInterests
  */
  
  /**Adds a LP into LPs.*/
  public void addLP(LogicalProcess lp)
  {
      schedulingLoop.LPs.add(lp);
  }//end addLP
  
  
  /**Returns the id of the PE.*/
  public int getID()
  {
      return myID;
  }
  
  /**Sets the id for the PE (dummy method).
    *@param theID int representing the value that the PE's id will be set to.
	*/
  public void setID(int theID)
  {
      myID=theID; 
  }
 


}//end class PE
